#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Vasya Svintsov <v.svintsov@techokert.ru>

from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formataddr
from typing import Union, Optional, List

from bs4 import BeautifulSoup
from http_tools.attached_file import AttachedFile
from http_tools.charset import Charset


def generate_message(source_email_account: str,
                     to_recipients: Union[str, List[str]],
                     subject: str,
                     body: str,
                     *,
                     source_email_account_name: Optional[str] = None,
                     attachments: Optional[List[AttachedFile]] = None,
                     cc_recipients: Optional[Union[str, List[str]]] = None,
                     bcc_recipients: Optional[Union[str, List[str]]] = None,
                     message_id: Optional[str] = None,
                     in_reply_to: Optional[str] = None,
                     references: Optional[str] = None
                     ) -> MIMEMultipart:
    """
    Creates a multipart/* type message.

    :param source_email_account: The source email account.
    :param to_recipients: The destination email accounts.
    :param subject: The subject of the email.
    :param body: The plain text/HTML version of the body of the email.
    :param source_email_account_name: The name of the source email account.
    :param attachments: The attachment part of the body of the email.
    :param cc_recipients: The list of recipients on the 'CC'.
    :param bcc_recipients: The list of recipients on the 'BCC'.
    :param message_id: string a single unique message identifier RFC 2822.
    :param in_reply_to: string field contain one or more unique message identifiers, optionally separated by CFWS RFC 2822.
    :param references: same as in_reply_to
    """

    if source_email_account_name is not None:
        source_email_account = formataddr((source_email_account_name, source_email_account), Charset.UTF8)

    message = MIMEMultipart('mixed')
    message['From'] = source_email_account
    message['Subject'] = subject

    to_recipients = [to_recipients] if not isinstance(to_recipients, list) else to_recipients
    message['To'] = ", ".join(to_recipients)
    if cc_recipients is not None:
        cc_recipients = [cc_recipients] if not isinstance(cc_recipients, list) else cc_recipients
        message['Cc'] = ", ".join(cc_recipients)
    if bcc_recipients is not None:
        bcc_recipients = [bcc_recipients] if not isinstance(bcc_recipients, list) else bcc_recipients
        message['Bcc'] = ", ".join(bcc_recipients)
    if message_id is not None:
        message['Message-ID'] = message_id
    if in_reply_to is not None:
        message['In-Reply-To'] = in_reply_to
    if references is not None:
        message['References'] = references

    message_body = MIMEMultipart('alternative')
    message_body.attach(MIMEText(BeautifulSoup(body, "html.parser").get_text(), 'plain', Charset.UTF8))
    message_body.attach(MIMEText(body, 'html', Charset.UTF8))
    message.attach(message_body)

    for attachment in attachments if attachments is not None else []:
        part = MIMEApplication(attachment.content)
        part.add_header("Content-Disposition", "attachment", filename=(Charset.UTF8, '', attachment.filename))
        message.attach(part)

    return message



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Vasya Svintsov <v.svintsov@techokert.ru>

import logging
from dataclasses import dataclass
from io import BytesIO
from typing import Union, List, Optional
from asyncio import TimeoutError
from aiohttp import FormData, ClientConnectorError
from aiohttp.web_exceptions import HTTPError
from http_tools.attached_file import AttachedFile
from http_tools.http_server_connector import HttpServerConnector

from http_tools.mime_types import ContentType

from ..email_client import EmailClient
from ...exceptions import SendFailed
from ...sending_result import SendingResult


logger = logging.getLogger(__name__)


class HTTPEmailClient(EmailClient):
    URL = "/v1/email/send"

    @dataclass
    class Config(HttpServerConnector.Config):
        ...

    @dataclass
    class Context(HttpServerConnector.Context):
        ...

    def __init__(self, config: Config, context: Context) -> None:
        self._http_connector = HttpServerConnector(config, context)

    async def send(self,
                   to_recipients: Union[str, List[str]],
                   subject: str,
                   body: str,
                   source_email_account: Optional[str] = None,
                   *,
                   source_email_account_name: Optional[str] = None,
                   attachments: Optional[List[AttachedFile]] = None,
                   cc_recipients: Optional[Union[str, List[str]]] = None,
                   bcc_recipients: Optional[Union[str, List[str]]] = None
                   ) -> SendingResult:
        form_data = FormData()

        values = {"subject": subject, "body": body, "to_recipients": to_recipients}
        if source_email_account is not None:
            values["source_email_account"] = source_email_account

        if source_email_account_name is not None:
            values["source_email_account_name"] = source_email_account_name

        if cc_recipients is not None:
            values["cc_recipients"] = cc_recipients

        if bcc_recipients is not None:
            values["bcc_recipients"] = bcc_recipients

        form_data.add_field("values", values, content_type=ContentType.Json.value)

        for item, attachment in enumerate(attachments, 1) if attachments is not None else []:
            form_data.add_field(f'file_{item}', BytesIO(attachment.content), filename=attachment.filename)

        try:
            payload = await self._http_connector.post(path=self.URL, payload=form_data)
            sending_result = SendingResult(**payload["result"])
            if not sending_result.status:
                raise SendFailed("unsuccessful send in gateway")
            logger.info(f"Email sent from {source_email_account} to {to_recipients} successfully.")
            result = sending_result
        except (HTTPError, TimeoutError, SendFailed, ClientConnectorError) as er:
            logger.error(f"Email sent from {source_email_account} to {to_recipients} failed, because {repr(er)}.")
            result = SendingResult(False)
        return result



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Vasya Svintsov <v.svintsov@techokert.ru>

from abc import ABC, abstractmethod
from typing import Union, List, Optional

from http_tools.attached_file import AttachedFile

from ..sending_result import SendingResult


class EmailClient(ABC):
    @abstractmethod
    async def send(self,
                   to_recipients: Union[str, List[str]],
                   subject: str,
                   body: str,
                   source_email_account: str,
                   *,
                   source_email_account_name: Optional[str] = None,
                   attachments: Optional[List[AttachedFile]] = None,
                   cc_recipients: Optional[Union[str, List[str]]] = None,
                   bcc_recipients: Optional[Union[str, List[str]]] = None
                   ) -> SendingResult:
        """
        Send the email.

        Note: source_email_account must be verified.

        :param to_recipients: The destination email accounts.
        :param subject: The subject of the email.
        :param body: The plain text/HTML version of the body of the email.
        :param source_email_account: The source email account.
        :param source_email_account_name: The name of the source email account.
        :param attachments: The attachment part of the body of the email.
        :param cc_recipients: The list of recipients on the 'CC'.
        :param bcc_recipients: The list of recipients on the 'BCC'.

        """
        ...



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Vasya Svintsov <v.svintsov@techokert.ru>
import json
from functools import partial
from typing import Any


def convert_to_list(value: Any, element_type: type) -> list:
    if isinstance(value, str):
        try:
            value = json.loads(value)
        except ValueError:
            ...
    if not isinstance(value, list):
        value = [value]
    value = [element_type(val) for val in value]
    return value


convert_to_list_of_string = partial(convert_to_list, element_type=str)
convert_to_list_of_int = partial(convert_to_list, element_type=int)



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Vasya Svintsov <v.svintsov@techokert.ru>

import logging
from dataclasses import dataclass
from typing import Union, List, Optional

from aiohttp import ClientSession
from dict_caster import DictCaster, Item
from furl import furl
from http_tools import HttpServer, Answer, WrappedAnswer
from http_tools.attached_file import AttachedFile

from http_tools.request import IncomingRequest


from .email_clients.amazon_ses_email_client.amazon_ses_email_client import AmazonSESEmailClient
from .email_clients.http_email_client.http_email_client import HTTPEmailClient
from .email_clients.smtp_email_client.smtp_email_client import SMTPEmailClient
from .sending_result import SendingResult


logger = logging.getLogger(__name__)


class EmailSender:
    @dataclass
    class _DefaultConfig:
        default_from_name: str = None

    @dataclass
    class _BaseConfig:
        default_from_email: str

    @dataclass
    class SMTPConfig(_DefaultConfig, SMTPEmailClient.Config, _BaseConfig):
        ...

    @dataclass
    class AmazonSESConfig(_DefaultConfig, AmazonSESEmailClient.Config, _BaseConfig):
        ...

    @dataclass
    class HTTPConfig(_DefaultConfig, HTTPEmailClient.Config):
        default_from_email: str = None

    @dataclass
    class Context(HTTPEmailClient.Context):
        session: ClientSession = None
        http_server: HttpServer = None

    def __init__(self,
                 config: Union[SMTPConfig, AmazonSESConfig, HTTPConfig],
                 context: Context,
                 path_prefix: Optional[str] = None,
                 allow_accept_http_requests: bool = False
                 ):
        self.config = config

        if isinstance(self.config, self.SMTPConfig):
            self._email_client = SMTPEmailClient(self.config)
        elif isinstance(self.config, self.AmazonSESConfig):
            self._email_client = AmazonSESEmailClient(self.config)
        elif isinstance(self.config, self.HTTPConfig):
            self._email_client = HTTPEmailClient(self.config, context)
        else:
            raise TypeError("InappropriateConfig")

        if context.http_server is not None and allow_accept_http_requests:
            path = furl(path_prefix).add(path=HTTPEmailClient.URL).url
            context.http_server.register_handler(path, self._send)

        logger.info(f"{type(self).__name__} init. EmailClientType: {type(self._email_client)}")

    async def send(self,
                   to_recipients: Union[str, List[str]],
                   subject: str,
                   body: str,
                   source_email_account: Optional[str] = None,
                   *,
                   source_email_account_name: Optional[str] = None,
                   attachments: Optional[List[AttachedFile]] = None,
                   cc_recipients: Optional[Union[str, List[str]]] = None,
                   bcc_recipients: Optional[Union[str, List[str]]] = None
                   ) -> SendingResult:
        """
        Send the email.

        Note: source_email_account must be verified.

        :param to_recipients: The destination email accounts.
        :param subject: The subject of the email.
        :param body: The plain text/HTML version of the body of the email.
        :param source_email_account: The source email account.
        :param source_email_account_name: The name of the source email account.
        :param attachments: The attachment part of the body of the email.
        :param cc_recipients: The list of recipients on the 'CC'.
        :param bcc_recipients: The list of recipients on the 'BCC'.

        """
        if source_email_account is None:
            source_email_account = self.config.default_from_email
        if source_email_account_name is None:
            source_email_account_name = self.config.default_from_name

        result = await self._email_client.send(
            to_recipients, subject, body, source_email_account,
            source_email_account_name=source_email_account_name,
            attachments=attachments,
            cc_recipients=cc_recipients,
            bcc_recipients=bcc_recipients
        )
        return result

    async def _send(self, request: IncomingRequest) -> Answer:
        values, attached_files = DictCaster(
            Item("values", dict),
            Item("attached_files", None, optional=True),
        ).cast_and_return(request.key_value_arguments)

        send_params = DictCaster(
            Item("subject", str),
            Item("body", str),
            Item("to_recipients", list),
            Item("source_email_account", str, optional=True),
            Item("source_email_account_name", str, optional=True),
            Item("cc_recipients", list, optional=True),
            Item("bcc_recipients", list, optional=True),
        ).cast(values)

        return WrappedAnswer(await self.send(**send_params, attachments=attached_files))



#  Copyright (C) 2024
#  ABM JSC, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Albakov Ruslan <r.albakov@abm-jsc.ru>
from .email_clients.amazon_ses_email_client.amazon_ses_email_client import AmazonSESEmailClient
from .email_clients.http_email_client.http_email_client import HTTPEmailClient
from .email_clients.smtp_email_client.smtp_email_client import SMTPEmailClient
from .email_sender import EmailSender
from .sending_result import SendingResult

__all__ = (AmazonSESEmailClient, HTTPEmailClient, SMTPEmailClient, EmailSender, SendingResult)



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Vasya Svintsov <v.svintsov@techokert.ru>

from dataclasses import dataclass

from aioboto3 import client
from aiobotocore.session import ClientCreatorContext


class AmazonSESConnector:
    @dataclass
    class Config:
        access_key: str
        secret_key: str
        region_name: str
        endpoint_url: str = None
        use_ssl: bool = True

    def __init__(self, config: Config):
        self._config = config

    def get_client_creator(self) -> ClientCreatorContext:
        client_session = client(
            service_name="ses",
            region_name=self._config.region_name,
            aws_access_key_id=self._config.access_key,
            aws_secret_access_key=self._config.secret_key,
            endpoint_url=self._config.endpoint_url)
        return client_session



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Vasya Svintsov <v.svintsov@techokert.ru>
import logging
import ssl
from dataclasses import dataclass
from enum import Enum
from typing import Union, List, Optional

from email.utils import make_msgid
import aiosmtplib
from http_tools.attached_file import AttachedFile

from ..email_client import EmailClient
from ...sending_result import SendingResult
from ...tools.generate_email_message import generate_message


logger = logging.getLogger(__name__)


class ConnectionType(str, Enum):
    SSL = "SSL"
    TLS = "TLS"
    STARTTLS = "STARTTLS"


class SMTPEmailClient(EmailClient):
    @dataclass
    class Config:
        server_hostname: str
        server_ssl_port: int
        username: str
        password: str
        connection_type: str = "TLS"  # SSL | TLS | STARTTLS
        domain_name: str | None = None

    def __init__(self, config: Config):
        self._config = config

        connection_type = self._config.connection_type.upper()
        self._use_tls = True if connection_type == ConnectionType.TLS else False
        self._start_tls = True if connection_type == ConnectionType.STARTTLS else False
        self._tls_context = ssl._create_stdlib_context() if connection_type != ConnectionType.SSL else None

    async def send(self,
                   to_recipients: Union[str, List[str]],
                   subject: str,
                   body: str,
                   source_email_account: str,
                   *,
                   source_email_account_name: Optional[str] = None,
                   attachments: Optional[List[AttachedFile]] = None,
                   cc_recipients: Optional[Union[str, List[str]]] = None,
                   bcc_recipients: Optional[Union[str, List[str]]] = None,
                   in_reply_to: Optional[str] = None,
                   references: Optional[str] = None
                   ) -> SendingResult:
        message_id = make_msgid(domain=self._config.domain_name)
        message = generate_message(source_email_account, to_recipients, subject, body,
                                   source_email_account_name=source_email_account_name,
                                   attachments=attachments,
                                   cc_recipients=cc_recipients,
                                   bcc_recipients=bcc_recipients,
                                   message_id=message_id,
                                   in_reply_to=in_reply_to,
                                   references=references)
        try:
            await aiosmtplib.send(
                message,
                hostname=self._config.server_hostname,
                port=self._config.server_ssl_port,
                use_tls=self._use_tls,
                start_tls=self._start_tls,
                username=self._config.username,
                password=self._config.password,
                tls_context=self._tls_context
            )
            logger.info(f"Email sent from {source_email_account} to {to_recipients} successfully.")
            result = SendingResult(True, message_id)
        except aiosmtplib.errors.SMTPException as er:
            logger.error(f"Email sent from {source_email_account} to {to_recipients} failed, because {repr(er)}.")
            result = SendingResult(False, message_id)
        return result



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Vasya Svintsov <v.svintsov@techokert.ru>

import logging
from dataclasses import dataclass
from typing import Union, List, Optional

from botocore.exceptions import ClientError
from http_tools.attached_file import AttachedFile


from .amazon_ses_connector import AmazonSESConnector
from ..email_client import EmailClient
from ...sending_result import SendingResult

from ...tools.generate_email_message import generate_message


logger = logging.getLogger(__name__)


class AmazonSESEmailClient(EmailClient):
    @dataclass
    class Config(AmazonSESConnector.Config):
        ...

    def __init__(self, config: Config):
        self._ses_connector = AmazonSESConnector(config)

    async def send(self,
                   to_recipients: Union[str, List[str]],
                   subject: str,
                   body: str,
                   source_email_account: str,
                   *,
                   source_email_account_name: Optional[str] = None,
                   attachments: Optional[List[AttachedFile]] = None,
                   cc_recipients: Optional[Union[str, List[str]]] = None,
                   bcc_recipients: Optional[Union[str, List[str]]] = None
                   ) -> SendingResult:
        message = generate_message(source_email_account, to_recipients, subject, body,
                                   source_email_account_name=source_email_account_name,
                                   attachments=attachments,
                                   cc_recipients=cc_recipients,
                                   bcc_recipients=bcc_recipients)
        send_args = {
            'Source': message["From"],
            "RawMessage": {'Data': message.as_string()},
            'Destinations': [to_recipients] if not isinstance(to_recipients, list) else to_recipients
        }
        try:
            async with self._ses_connector.get_client_creator() as client_session:
                response = await client_session.send_raw_email(**send_args)
            message_id = response['MessageId']
            logger.info(f"Email({message_id}) sent from {source_email_account} to {to_recipients} successfully.")
            result = SendingResult(True, message_id)
        except ClientError as er:
            logger.error(f"Email sent from {source_email_account} to {to_recipients} failed, because {repr(er)}.")
            result = SendingResult(False)
        return result



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Vasya Svintsov <v.svintsov@techokert.ru>

from dataclasses import dataclass
from typing import Optional


@dataclass
class SendingResult:
    status: bool
    message_id: Optional[str] = None
